<?php
/*
	q2Packet Class by ViciouZ
	(c) 2010 Steve Cox, ViciouZ.co.uk
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

include_once("defines.php");

class q2Packet
{
	// Stored data
	private $packetPayload; // Packet data
	private $payloadSize; // Total size of packet
	private $streamLocation; // Current position in $this->packetPayload
	private $debugPrintAll; // print_r is called on all data
	
	private $dataHandlers; // Array of callbacks for each message type
	
	private $numMessages; // Number of messages parsed
	private $numReconnects; // Number of reconnects (indicate map changes)
	
	private $keepParsing = true; // Control logic
	private $errorString; // Current error message, should only be set if an error has occured.
	
	// Constructor function
	function __construct($packetdata = "", $debug = false)
	{	
		$this->payloadSize = strlen($packetdata);
		$this->packetPayload = $packetdata;
		$this->debugPrintAll = $debug;
	}
	
	// Replaces the packet data and resets the stream location without having to instantiate a new class
	function NewPacket($packetdata)
	{
		$this->payloadSize = strlen($packetdata);
		$this->packetPayload = $packetdata;
		$this->streamLocation = 0;
	}
	
	// Read a signed char from the stream
	private function ReadChar()
	{
		$data = substr($this->packetPayload,$this->streamLocation,1);
		$arr = unpack("c1char",$data);
		$this->streamLocation += 1;
		return $arr['char'];
	}
	
	// Read an unsigned char (byte) from the stream
	private function ReadByte()
	{
		$data = substr($this->packetPayload,$this->streamLocation,1);
		$arr = unpack("C1byte",$data);
		$this->streamLocation += 1;
		return $arr['byte'];
	}
	
	// Read a signed short int (2bytes) from the stream
	function ReadShort()
	{
		$data = substr($this->packetPayload,$this->streamLocation,2);
		$arr = unpack("s1short",$data);
		$this->streamLocation += 2;
		return $arr['short'];
	}
	
	// Read a signed long int (4bytes) from the stream
	private function ReadLong()
	{
		$data = substr($this->packetPayload,$this->streamLocation,4);
		$arr = unpack("l1long",$data);
		$this->streamLocation += 4;
		return $arr['long'];
	}
	
	// Replace the 5.3.0 dependant "before_needle" parameter
	private function strstr2($haystack,$needle)
	{
		return array_shift(explode($needle,$haystack,2));
	}
	
	// Read a string up until the next null character in the stream
	private function ReadString()
	{
		$data = substr($this->packetPayload,$this->streamLocation);
		$string = $this->strstr2($data,chr(0));
		$this->streamLocation += strlen($string)+1;
		return $string;
	}
	
	// Read 3 shorts into an array (useful for some position data)
	private function Read3Shorts()
	{
		$data = substr($this->packetPayload,$this->streamLocation,6);
		$arr = unpack("s3",$data);
		$this->streamLocation += 6;
		return $arr;
	}
	
	// Read an angle stored in one byte
	private function ReadAngle()
	{
		return (1.40625*$this->ReadChar());
	}
	
	// Read an angle stored in 2 bytes
	private function ReadShortAngle()
	{
		return ((360/65536)*$this->ReadShort());
	}
	
	// Read 3 vectors stored as short ints
	private function ReadVector()
	{
		$arr = $this->Read3Shorts();
		foreach ($arr as &$value)
		{
			$value = $value / 8;
		}
		return $arr;
	}
	
	// Read a single co-ordinate stored as a short int
	private function ReadCoord()
	{
		return ($this->ReadShort()/8);
	}
	
	// Read data of a specified length from the stream
	private function ReadData($length)
	{
		$data = substr($this->packetPayload,$this->streamLocation,$length);
		$this->streamLocation += $length;
		return $data;
	}
	
	// Main function
	// $fastparse - bool - if true, where possible, don't read data if it's not going to be handled
	function ParsePacket($fastparse = false)
	{
		if($fastparse)
		{
			for($i=0;$i<=21;$i++)
			{
				if(!isset($this->dataHandlers[$i]))
					$fastArray[$i] = true;
				else
					$fastArray[$i] = false;
			}
		}
		else
			$param = false;
	
		while($this->keepParsing)
		{
			// Have we reached the end of the stream
			if($this->streamLocation >= $this->payloadSize)
				break;
		
			// Get the command type indentifier
			$svc = $this->ReadByte();
			if(!$svc)
				break;
				
			if($fastparse)
				$param = $fastArray[$svc];
				
			$res = null;
				
			switch($svc)
			{
				case svc_bad:
					break;
					
				case svc_muzzleflash:
					$res = $this->ParseMuzzleFlash($param);
					break;
					
				case svc_muzzleflash2:
					$res = $this->ParseMuzzleFlash2($param);
					break;
					
				case svc_temp_entity:
					$res = $this->ParseTempEntity($param);
					break;
					
				case svc_layout:
					$res = $this->ParseLayout($param);
					break;
					
				case svc_inventory:
					$res = $this->ParseInventory($param);
					break;
					
				case svc_nop:
				case svc_disconnect:
					$this->keepParsing = false;
					break;
					
				case svc_reconnect:
					$this->numReconnects++;
					// Note - although these have no parsing functions they can still be hooked with AddHandler.
					break;
				
				case svc_sound:
					$res = $this->ParseSound($param);
					break;
					
				case svc_print:
					$res = $this->ParsePrint($param);
					break;
				
				case svc_stufftext:
					$res = $this->ParseStuffText($param);
					break;
				
				case svc_configstring:
					$res = $this->ParseConfigString($param);
					break;
				
				case svc_serverdata:
					$res = $this->ParseServerData($param);
					break;
				
				case svc_spawnbaseline:
					$res = $this->ParseSpawnBaseline($param);
					break;
				
				case svc_centerprint:
					$res = $this->ParseCenterPrint($param);
					break;
				
				case svc_download:
					// This probably shouldn't be here, skip rest of packet
					$this->streamLocation = $this->payloadSize;
					break;
					
				case svc_playerinfo:
					// Always second last - if nothing else is needed, we can skip the rest of the packet
					if(($param) && ($fastArray[svc_packetentities]))
					{
						$this->streamLocation = $this->payloadSize;
						break;
					}
					$res = $this->ParsePlayerInfo($param);
					break;
				
				case svc_packetentities:
					// Always the last command in a packet - if it's not needed, we can skip the rest
					if($param)
					{
						$this->streamLocation = $this->payloadSize;
						break;
					}
					$res = $this->ParsePacketEntities();
					break;
				
				case svc_deltapacketentities:
					// This isn't even used anymore I don't think... skip rest of packet
					$this->streamLocation = $this->payloadSize;
					break;
				
				case svc_frame:
					// Always third last - if nothing else is needed, we can skip the rest of the packet
					if(($param) && ($fastArray[svc_playerinfo]) && ($fastArray[svc_packetentities]))
					{
						$this->streamLocation = $this->payloadSize;
						break;
					}
					$res = $this->ParseFrame($param);
					break;
					
				default:
					$this->errorString = "Encountered unknown svc type: ".$svc;
					$this->keepParsing = false;
					break;
			}
			
			// Call any data handlers set for this svc code
			if(isset($this->dataHandlers[$svc]))
				call_user_func($this->dataHandlers[$svc],$res);
			
			// Debug printing
			if($this->debugPrintAll)
			{
				global $svc_to_names;
				print $svc_to_names[$svc]."\n";
				print_r($res);
				print "\n";
			}
			
			$this->numMessages++;
		}
		
		if(($this->keepParsing != true) && isset($this->errorString))
			return $this->errorString;
		else
			return true;
	}
	
	// Parses a MuzzleFlash message
	private function ParseMuzzleFlash($fast = false)
	{
		if($fast)
		{$this->streamLocation += 3; return; }
		
		
		$arr['entity'] = $this->ReadShort();
		$arr['weapon'] = $this->ReadByte();
		return $arr;
	}
	
	// Parses a MuzzleFlash (2nd type) message
	private function ParseMuzzleFlash2($fast = false)
	{
		if($fast)
			{$this->streamLocation += 3; return; }
		
		$arr['entity'] = $this->ReadShort();
		$arr['flash_num'] = $this->ReadByte();
		return $arr;
	}
	
	// Parse a TempEntity message
	private function ParseTempEntity($fast = false)
	{
		$arr['type'] = $this->ReadByte();
		switch($arr['type'])
		{
			case 10: // TE_SPLASH
			case 25: // TE_WELDING_SPARKS
			case 29: // TE_TUNNEL_SPARKS
			case 15: // TE_LASER_SPARKS - paintball splats
				if($fast)
					{$this->streamLocation += 9; return; }
				$arr['count'] = $this->ReadByte();
				$arr['position'] = $this->ReadVector();
				$arr['direction'] = $this->ReadByte();
				$arr['colour'] = $this->ReadByte();
				break;
			case 6: // TE_EXPLOSION2
			case 8: // TE_GRENADE_EXPLOSION
			case 18: // TE_GRENADE_EXPLOSION_WATER
			case 5: // TE_EXPLOSION1
			case 53: // TE_EXPLOSION1_BIG
			case 7: // TE_ROCKET_EXPLOSION
			case 17: // TE_ROCKET_EXPLOSION_WATER
			case 54: // TE_EXPLOSION1_NP
			case 20: // TE_BFG_EXPLOSION
			case 21: // TE_BFG_BIGEXPLOSION
			case 35: // TE_PLAIN_EXPLOSION
			case 48: // TE_TELEPORT_EFFECT
			case 49: // TE_DBALL_GOAL
			case 52: // TE_WIDOWSPLASH
			case 51: // TE_NUKEBLAST
			case 45: // TE_CHAINFIST_SMOKE - paintball smoke effects
				if($fast)
					{$this->streamLocation += 6; return; }
				$arr['position'] = $this->ReadVector();
				break;
			case 0: // TE_GUNSHOT
			case 9: // TE_SPARKS
			case 14: // TE_BULLET_SPARKS
			case 12: // TE_SCREEN_SPARKS
			case 13: // TE_SHIELD_SPARKS
			case 4: // TE_SHOTGUN
			case 2: // TE_BLASTER
			case 30: // TE_BLASTER2
			case 55: // TE_FLECHETTE
				if($fast)
					{$this->streamLocation += 7; return; }
				$arr['position'] = $this->ReadVector();
				$arr['direction'] = $this->ReadByte();
				break;
			case 19: // TE_MEDIC_CABLE_ATTACK
				if($fast)
					{$this->streamLocation += 14; return; }
				$arr['entity'] = $this->ReadShort();
			case 3: // TE_RAILTRAIL
			case 23: // TE_BFG_LASER
			case 11: // TE_BUBBLETRAIL
			case 41: // TE_BUBBLETRAIL2
			case 34: // TE_DEBUGTRAIL
				if($fast)
					{$this->streamLocation += 12; return; }
				$arr['position1'] = $this->ReadVector();
				$arr['position2'] = $this->ReadVector();
				break;
			case 36: // TE_FLASHLIGHT
				if($fast)
					{$this->streamLocation += 8; return; }
				$arr['position'] = $this->ReadVector();
				$arr['entity'] = $this->ReadShort();
				break;
			case 37: // TE_FORCEWALL
				if($fast)
					{$this->streamLocation += 13; return; }
				$arr['position1'] = $this->ReadVector();
				$arr['position2'] = $this->ReadVector();
				$arr['colour'] = $this->ReadByte();
				break;
			case 40: // TE_STEAM
				$arr['id'] = $this->ReadShort();
				$arr['count'] = $this->ReadByte();
				$arr['position'] = $this->ReadVector();
				$arr['direction'] = $this->ReadByte();
				$arr['colour'] = $this->ReadByte();
				$arr['magnitude'] = $this->ReadShort();
				if($arr['id'] != -1)
					$arr['duration'] = $this->ReadLong();
				break;
			case 50: // TE_WIDOWBEAMOUT
				if($fast)
					{$this->streamLocation += 8; return; }
				$arr['id'] = $this->ReadShort();
				$arr['position'] = $this->ReadVector();
				break;
			default:
				$this->errorString = "q2Packet: Unknown tempentity type: ".$arr['type'];
				$this->keepParsing = false;
				return false;
		}		
		return $arr;
	}
	
	// Parse the player's inventory message
	private function ParseInventory($fast = false)
	{
		if($fast)
			{$this->streamLocation += 512; return; }
			
		for($i=0;$i<256;$i++)
			$arr[$i] = $this->ReadShort();
		return $arr;
	}
	
	// Parse a sound message (uses bits)
	// TODO - optimise this?
	private function ParseSound($fast = false)
	{
		$bits = $this->ReadByte();
		$arr['soundindex'] = $this->ReadByte();
		
		if($bits & SND_VOLUME)
			$arr['volume'] = $this->ReadByte()/255;
		if($bits & SND_ATTENUATION)
			$arr['attenuation'] = $this->ReadByte()/64;
		if($bits & SND_OFFSET)
			$arr['offset'] = $this->ReadByte()/1000;
		if($bits & SND_ENT)
		{
			$arr['channel'] = $this->ReadShort();
			$arr['entity'] = ($arr['channel']>>3);
			$arr['channel'] &= 7;
		}
		if($bits & SND_POS)
			$arr['position'] = $this->ReadVector();
			
		return $arr;
	}
	
	// Parse a layout string message
	private function ParseLayout($fast = false)
	{
		$layoutstring = $this->ReadString();
		return $layoutstring;
	}
	
	// Parse a print message
	private function ParsePrint($fast = false)
	{
		$arr['level'] = $this->ReadByte();
		$arr['string'] = $this->ReadString();
		return $arr;
	}
	
	// Parse a stuffed command
	private function ParseStuffText($fast = false)
	{
		$command = $this->ReadString();
		return $command;
	}
	
	// Parse a configstring message
	private function ParseConfigString($fast = false)
	{
		$arr['index'] = $this->ReadShort();
		$arr['string'] = $this->ReadString();
		return $arr;
	}
	
	// Parse the server data messge (first message of first packet)
	private function ParseServerData($fast = false)
	{		
		$arr['protocol'] = $this->ReadLong();
		$arr['servercount'] = $this->ReadLong();
		$arr['attractloop'] = $this->ReadByte();
		$arr['gamedir'] = $this->ReadString();
		$arr['player_entity_num'] = $this->ReadShort();
		$arr['maptitle'] = $this->ReadString();	
		return $arr;
	}
	
	// Parse a centerprint message
	private function ParseCenterPrint($fast = false)
	{
		$message = $this->ReadString();
		return $message;
	}
	
	// Parse the player info - uses bits
	private function ParsePlayerInfo($fast = false)
	{
		$flags = $this->ReadShort();
		
		if($flags & PS_M_TYPE)
			$arr['pm_type'] = $this->ReadByte();
		if($flags & PS_M_ORIGIN)
			$arr['pmove_origin'] = $this->ReadVector();
		if($flags & PS_M_VELOCITY)
			$arr['pmove_velocity'] = $this->Read3Shorts();
		if($flags & PS_M_TIME)
			$arr['pmove_time'] = $this->ReadByte();
		if($flags & PS_M_FLAGS)
			$arr['pmove_flags'] = $this->ReadByte();
		if($flags & PS_M_GRAVITY)
			$arr['pmove_gravity'] = $this->ReadShort();
		if($flags & PS_M_DELTA_ANGLES)
		{
			for($i=1;$i<=3;$i++)
				$arr['pmove_delta_angles'][$i] = $this->ReadShortAngle();
		}
		if($flags & PS_VIEWOFFSET)
		{
			for($i=1;$i<=3;$i++)
				$arr['viewoffset'][$i] = ($this->ReadChar()/4);
		}
		if($flags & PS_VIEWANGLES)
		{
			for($i=1;$i<=3;$i++)
				$arr['viewangles'][$i] = $this->ReadShortAngle();
		}
		if($flags & PS_KICKANGLES)
		{
			for($i=1;$i<=3;$i++)
				$arr['kickangles'][$i] = ($this->ReadChar()/4);
		}
		if($flags & PS_WEAPONINDEX)
			$arr['gunindex'] = $this->ReadByte();
		if($flags & PS_WEAPONFRAME)
		{
			$arr['gunframe'] = $this->ReadByte();
			for($i=1;$i<=3;$i++)
				$arr['gunoffset'][$i] = ($this->ReadChar()/4);
			for($i=1;$i<=3;$i++)
				$arr['gunangles'][$i] = ($this->ReadChar()/4);
		}
		if($flags & PS_BLEND)
		{
			for($i=1;$i<=4;$i++)
				$arr['blend'][$i] = ($this->ReadByte()/255);
		}
		if($flags & PS_FOV)
			$arr['fov'] = $this->ReadByte();
		if($flags & PS_RDFLAGS)
			$arr['rdflags'] = $this->ReadByte();
			
		$arr['statsbits'] = $this->ReadLong();
		for($i=0;$i<32;$i++)
		{
			if($arr['statsbits'] & (1<<$i))
				$arr['stats'][$i] = $this->ReadShort();
		}
		return $arr;
	}
	
	// Loop parse the entities at the end of the packet
	private function ParsePacketEntities($fast = false)
	{
		while($this->streamLocation < $this->payloadSize)
		{
			$array = $this->ParseEntityBits();
			
			if($array['entity'] == 0)
				break;
			
			$arr[$array['entity']] = $this->ParseDeltaEntity($array);
		}
		return $arr;
	}
	
	// Parse the "frame" message
	private function ParseFrame($fast = false)
	{
		$arr['serverframe'] = $this->ReadLong();
		$arr['deltaframe'] = $this->ReadLong();
		$this->ReadByte(); // junk byte from protocol version 26
		
		if($arr['deltaframe'] <= 0)
			$arr['compressed'] = false;
		else
			$arr['compressed'] = true;
			
		$arr['areabitslength'] = $this->ReadByte();
		$arr['areabits'] = $this->ReadData($arr['areabitslength']);
		
		return $arr;
	}
	
	// Check and read the bits which indicate delta compression
	private function ParseEntityBits()
	{
		$bits = $this->ReadByte();
		if($bits & U_MOREBITS1)
		{
			$newbits = $this->ReadByte();
			$bits |= ($newbits<<8);
		}
		if($bits & U_MOREBITS2)
		{
			$newbits = $this->ReadByte();
			$bits |= ($newbits<<16);
		}
		if($bits & U_MOREBITS3)
		{
			$newbits = $this->ReadByte();
			$bits |= ($newbits<<24);
		}
		
		// the entity number is stored as 1 or 2 bytes
		if($bits & U_NUMBER16)
			$arr['entity'] = $this->ReadShort();
		else
			$arr['entity'] = $this->ReadByte();
			
		$arr['bits'] = $bits;
			
		return $arr;
	}
	
	// Parse an entity message with delta compression - uses bits
	private function ParseDeltaEntity($arr)
	{
		$bits = $arr['bits'];
		if($bits & U_MODEL)
			$arr['modelindex'] = $this->ReadByte();
		if($bits & U_MODEL2)
			$arr['modelindex2'] = $this->ReadByte();
		if($bits & U_MODEL3)
			$arr['modelindex3'] = $this->ReadByte();
		if($bits & U_MODEL4)
			$arr['modelindex4'] = $this->ReadByte();
			
		if($bits & U_FRAME8)
			$arr['frame'] = $this->ReadByte();
		else if($bits & U_FRAME16)
			$arr['frame'] = $this->ReadShort();
		
		if(($bits & U_SKIN8) && ($bits & U_SKIN16))
			$arr['skinnum'] = $this->ReadLong();
		else if($bits & U_SKIN16)
			$arr['skinnum'] = $this->ReadShort();
		else if($bits & U_SKIN8)
			$arr['skinnum'] = $this->ReadByte();
			
		if(($bits & (U_EFFECTS8|U_EFFECTS16)) == (U_EFFECTS8|U_EFFECTS16))
			$arr['effects'] = $this->ReadLong();
		else if($bits & (U_EFFECTS16))
			$arr['effects'] = $this->ReadShort();
		else if($bits & (U_EFFECTS8))
			$arr['effects'] = $this->ReadByte();
			
		if(($bits & (U_RENDERFX8|U_RENDERFX16)) == (U_RENDERFX8|U_RENDERFX16))
			$arr['renderfx'] = $this->ReadLong();
		else if($bits & (U_RENDERFX16))
			$arr['renderfx'] = $this->ReadShort();
		else if($bits & (U_RENDERFX8))
			$arr['renderfx'] = $this->ReadByte();
			
		if($bits & U_ORIGIN1)
			$arr['origin'][0] = $this->ReadCoord();
		if($bits & U_ORIGIN2)
			$arr['origin'][1] = $this->ReadCoord();
		if($bits & U_ORIGIN3)
			$arr['origin'][2] = $this->ReadCoord();
			
		if($bits & U_ANGLE1)
			$arr['angle'][0] = $this->ReadAngle();
		if($bits & U_ANGLE2)
			$arr['angle'][1] = $this->ReadAngle();
		if($bits & U_ANGLE3)
			$arr['angle'][2] = $this->ReadAngle();

		if($bits & U_OLDORIGIN)
			$arr['old_origin'] = $this->ReadVector();
		if($bits & U_SOUND)
			$arr['sound'] = $this->ReadByte();
		if($bits & U_EVENT)
			$arr['event'] = $this->ReadByte();
		if($bits & U_SOLID)
			$arr['solid'] = $this->ReadShort();
	
		return $arr;
	}
	
	// Parse a baseline message (initial entity data)
	private function ParseSpawnBaseline($fast = false)
	{
		$array = $this->ParseEntityBits();
		$arr = $this->ParseDeltaEntity($array);
		
		return $arr;
	}
	
	// Adds a data handler callback to $this->dataHandlers
	// These callbacks are called with an array of data after a message is parsed
	// One callback can be set for each svc code
	function AddHandler($svc_number, $function_name)
	{
		if(is_callable($function_name))
			$this->dataHandlers[$svc_number] = $function_name;
		else
			echo "q2Packet::AddHandler - Function \"".$function_name."\" not callable!\n";
	}
	
	// Removes a data handler from the $this->dataHandlers array
	function RemoveHandler($svc_number)
	{
		global $svc_to_names;
	
		if(isset($this->dataHandlers[$svc_number]))
			unset($this->dataHandlers[$svc_number]);
		else
			echo "q2Packet::RemoveHandler - Handler for ".$svc_to_names[$svc_number]." not found!\n";
	}
	
	// Object use stats
	function GetNumMessages()
	{
		return array("messages" => $this->numMessages, "reconnects" => $this->numReconnects);
	}
}
?>